SystemOrganization addCategory: #'Neo-Unicode-Normalization'!

Object subclass: #NeoNormalizedStream
	instanceVariableNames: 'form sourceStream currentSequence seqIndex nextSequence'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Unicode-Normalization'!

!NeoNormalizedStream commentStamp: 'HenrikSperreJohansen 12/14/2015 18:55' prior: 0!
A normalization stream returns a sequence of characters whose code points are in a normalized form.
Used for sorting and its ilk, its task is efficiently returning the next code point in the normalized string;
that way, we can construct CollateElements only for characters relevant to the sort, rather than the entire string.!

!NeoNormalizedStream methodsFor: 'as yet unclassified' stamp: 'HenrikSperreJohansen 12/14/2015 19:10'!
next
	seqIndex := (seqIndex \\ currentSequence  size) + 1.
	seqIndex = 1 ifTrue: [currentSequence := form nextFrom: sourceStream].
	^currentSequence at: seqIndex! !

!NeoNormalizedStream methodsFor: 'as yet unclassified' stamp: 'HenrikSperreJohansen 12/14/2015 15:59'!
peek
	! !

Object subclass: #NormalizationForm
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Unicode-Normalization'!

!NormalizationForm commentStamp: 'HenrikSperreJohansen 12/14/2015 19:24' prior: 0!
When two String are in  normalized form, if the strings are equivalent, their code point sequences will also be the same.

The decomposed forms both follow the same steps:
- Recursively map next code point to decomposed codePoint sequences.
- Add canonically combining characters.
- Sort combining code points 

Note; the elements returned by nextFrom: are not the same as abstract characters; already decomposed characters will be returned as 2 1-code point elements, not a single 2-code point element.
For sorting; converting to normalized form, etc, we don't really care about this, as long as the resulting code point sequence is in properly decomposed order.!

NormalizationForm subclass: #NFD
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Unicode-Normalization'!

!NFD commentStamp: 'HenrikSperreJohansen 12/14/2015 18:39' prior: 0!
NFD is the Canonical Decomposition normalization form.
All codepoints that have an canonical decomposition will be represented using the decomposed sequence.

!

!NFD methodsFor: 'private' stamp: 'HenrikSperreJohansen 12/14/2015 17:46'!
addCanonicalMapping: aCodePoint to: sequence
	| data |
	data := NeoUnicodeCharacterData forCodePoint: aCodePoint ifAbsent: nil.
	
	(data notNil and: [data hasCanonicalDecomposition ])
		ifTrue: [ data decompositionMapping do: [ :each | self addCanonicalMapping: each to: sequence] ] 
		ifFalse: [ sequence nextPut: (Character value: aCodePoint) ]! !

!NFD methodsFor: 'private' stamp: 'HenrikSperreJohansen 12/14/2015 18:45'!
addMapping: aCodePoint to: sequence
	^self addCanonicalMapping: aCodePoint to: sequence! !

NormalizationForm subclass: #NFKD
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Unicode-Normalization'!

!NFKD commentStamp: 'HenrikSperreJohansen 12/14/2015 10:37' prior: 0!
NFKD is the Compatibility Decomposition normalization form.
All codepoints that have an equivalent compatability decomposition will be represented using the decomposed sequence!

!NFKD methodsFor: 'private' stamp: 'HenrikSperreJohansen 12/14/2015 17:41'!
addCompatibilityMapping: aCodePoint to: sequence
	| data |
	data := NeoUnicodeCharacterData forCodePoint: aCodePoint ifAbsent: nil.
	(data notNil and: [ data decompositionMapping notNil ])
		ifTrue:
			[ data decompositionMapping first isSymbol
				ifTrue:
					[ "Compatibility mapping" data decompositionMapping 
						allButFirstDo: [ :each |  self addCompatibilityMapping: each to: sequence ] ]
				ifFalse: [ "Canonical mapping" data decompositionMapping 
						do: [ :each | self addCompatibilityMapping: each to: sequence ] ]		]
		ifFalse: [ sequence nextPut: (Character value: aCodePoint) ]! !

!NFKD methodsFor: 'private' stamp: 'HenrikSperreJohansen 12/14/2015 18:46'!
addMapping: aCodePoint to: sequence
 ^self addCompatibilityMapping: aCodePoint to: sequence
! !

!NormalizationForm class methodsFor: 'conversion' stamp: 'HenrikSperreJohansen 12/14/2015 10:47'!
next: aNumber from: aCodePointStream toStream: aDestinationStream 
	"Read the next aNumber abstract character in my normalized form from aCodePointStream, and place them on aDestinationStream.
	Return the amount of abstract characters actually read"
	^self subclassResponsibility! !

!NormalizationForm class methodsFor: 'conversion' stamp: 'HenrikSperreJohansen 12/14/2015 10:53'!
next: aNumber putAll: aString startingAt: aStartIndex toStream: aDestinationStream 
	"Read the next aNumber abstract character in my normalized form from aCodePointStream, and place them on aDestinationStream.
	Return the amount of abstract characters actually read"
	^self next: aNumber ! !

!NormalizationForm class methodsFor: 'conversion' stamp: 'HenrikSperreJohansen 12/14/2015 10:42'!
nextFrom: aCodePointStream
	"Reads the next abstract character in my normalized form from aCodePointStream"
	^self subclassResponsibility! !

!NormalizationForm methodsFor: 'private' stamp: 'HenrikSperreJohansen 12/14/2015 18:45'!
addMapping: aCodePoint to: aSequenceStream
	^self subclassResponsibility! !

!NormalizationForm methodsFor: 'private' stamp: 'HenrikSperreJohansen 12/16/2015 13:39'!
addReorderableCodePointsFrom: aStream to: sequence.
	[|next data |
	(next := aStream peek) ifNil: [^self].	
	data := NeoUnicodeCharacterData forCodePoint: next codePoint ifAbsent: nil.
	data notNil and: [data canonicalCombiningClass > 0]] whileTrue: [ self addMapping: aStream next codePoint to: sequence]! !

!NormalizationForm methodsFor: 'private' stamp: 'HenrikSperreJohansen 12/14/2015 18:42'!
combiningClassOf: a smallerThan: b
	^(NeoUnicodeCharacterData forCodePoint: a codePoint) canonicalCombiningClass <= 
	(NeoUnicodeCharacterData forCodePoint: b codePoint) canonicalCombiningClass! !

!NormalizationForm methodsFor: 'streaming' stamp: 'HenrikSperreJohansen 12/16/2015 16:10'!
nextFrom: aStream
	|  sequence result sortRanges rangeStart currentMax|
	sequence := String new writeStream.
	self addMapping: aStream next codePoint to: sequence.
	self addReorderableCodePointsFrom: aStream to: sequence.
	"Merge sort is stable, points with same Canonical Combining class will remain in order."
	"Some compatability decompositions (for instance U+3300) contain combining characters surrounded by starting characters, make sure we only sort sequences of combining characters. Make it work, then make it pretty, haven't quite gotten to the pretty part yet."
	result := sequence contents.
	sortRanges := OrderedCollection new.
	rangeStart := currentMax := 0.
	result doWithIndex: [:char :index | |curr| 
		curr := (NeoUnicodeCharacterData forCodePoint: char codePoint ifAbsent: nil) ifNil: [0] ifNotNil: [:data | data ccc ].
		curr = 0 
			ifFalse: [ rangeStart = 0 ifTrue: [rangeStart := index].]
			ifTrue: [rangeStart > 0 ifTrue: [
					index - rangeStart > 1 ifTrue: [sortRanges add: (rangeStart to: index -1)].
					rangeStart := 0] ]].
	"Add range at end of sequence"
	(rangeStart > 0 and: [rangeStart < result size]) ifTrue: [ sortRanges add: (rangeStart to: result size) ].
	sortRanges do: [ :range | 
		result mergeSortFrom: range first to: range last  by: [:a :b | self combiningClassOf: a smallerThan: b ] ].
	^result! !

!NormalizationForm methodsFor: 'streaming' stamp: 'HenrikSperreJohansen 12/14/2015 16:09'!
peekFrom: aCodePointStream
	"Peeks the next abstract character sequence in my normalized form from aCodePointStream. Consumes 1-N characters from stream, returns 1-M characters"
	^self subclassResponsibility! !
